StellarUF = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceDB-2.0", "AceHook-2.1")
local self = StellarUF
local surf = AceLibrary("Surface-1.0")
self.surface = surf

local dew = AceLibrary("Dewdrop-2.0")

local _G = getfenv(0)
local db, opts
local tar, pla, white, su, partyvisible, s30, s40
local dead, online = UnitIsDeadOrGhost, UnitIsConnected
local incombat = InCombatLockdown
local friend = UnitIsFriend
local exist, visible = UnitExists, UnitIsVisible
local nparty, nraid = GetNumPartyMembers, GetNumRaidMembers

function StellarUF:OnInitialize()
	_,self.cls = UnitClass("player")
	self.units = { }
	su = self.units
	self:LoadVariables()
	self:InitSpellRange()
	
	-- maybe unnecessary, but shouldn't hurt
	self:ScheduleLeaveCombatAction("LoadFrames")
end

function StellarUF:LoadFrames()
	self:HideBlizzStuff()

	self:CreateUnitFrame("player", "full")
	self:CreateUnitFrame("target", "full", true)
	self:CreateUnitFrame("targettarget", "small")
	self:CreateUnitFrame("focus", "small")
	self:CreateUnitFrame("focustarget", "small", true)
	self:CreateUnitFrame("pet", "medium")
	
	tar = su.target
	pla = su.player

	-- only care to create party frames dynamically, too much trouble imo
	if nparty() > 0 or nraid() > 0 then
		self:CreateParty()
	else
		self:RegisterEvent("PARTY_MEMBERS_CHANGED", "CreateParty")
	end
	self:Hook("RaidOptionsFrame_UpdatePartyFrames", "RaidPartyUpdate", true)

	self:RegisterEvents()
end


-- update texture when user changes it
local function UpTexture()
	local t = surf:Fetch(db.texture)
	for u,f in pairs(su) do
		f:SetBackdrop({ bgFile=t, })
		f:SetBackdropColor(0, 0, 0, 0.2)
		f.hp.bar:SetTexture(t)
		f.hp:SetBackdrop({ bgFile = t, })
		f.hp:SetBackdropColor(0, self.hpcolor.g/3.5, 0, 0.6)
		if f.hp.backbar then
			f.hp.backbar:SetTexture(t)
		end
		f.mp.bar:SetTexture(t)
		f.mp:SetBackdrop({ bgFile = t, })
		f.mp:SetBackdropColor(0, 0, 0, 0.6)
		if f.mp.backbar then
			f.mp.backbar:SetTexture(t)
		end
	end
end

-- update scaling when user changes it
local function UpScale(type)
	local s = db.scale[type]
	for u,f in pairs(su) do
		if f.frametype == type and u ~= "focustarget" then
			f:SetScale(s)
		end
	end
end
function StellarUF:LoadVariables()
	-- health bar color
	self.hpcolor = { r=0, g=0.7, b=0, }
	
	self:RegisterDB("StellarUFDB")
	self:RegisterDefaults("profile", {
		pos = {
			player = { x = 21, y = -35, },
			target = { x = 230, y = -35, },
			pet = { x = 60, y = -85, },
			targettarget = { x = 430, y = -40, },
			party1 = { x = 16, y = -130, },
			party2 = { x = 16, y = -180, },
			party3 = { x = 16, y = -230, },
			party4 = { x = 16, y = -280, },
			partypet1 = { x = 0, y = 0, }, -- party pets have dummy vars, it's set to owner
			partypet2 = { x = 0, y = 0, },
			partypet3 = { x = 0, y = 0, },
			partypet4 = { x = 0, y = 0, },
			focus = { x = 430, y = -100, },
			focustarget = { x = 430, y = -100, },
		},
		scale = {
			full = 1.0,
			medium = 1.0,
			small = 1.0,
		},
		texture = "Charcoal Low",
		showmodels = true, 
		deficit = false, 
	})
	db = self.db.profile
	
	surf:Register("Flatsmooth", "Interface\\AddOns\\StellarUF\\images\\flatsmooth")
	surf:Register("Charcoal Low", "Interface\\AddOns\\StellarUF\\images\\charcoallow")
	surf:Register("Steel", "Interface\\AddOns\\StellarUF\\images\\steel")
	self:RegisterEvent("Surface_SetGlobal", UpTexture)

	opts = {
		type = "group",
		args = {
			header = {
				type = "header",
				name = "StellarUF",
				order = 1,
			},
			texture = {
				type = "text",
				name = "Bar Texture",
				desc = "Statusbar textures",
				set = function(v) db.texture = v UpTexture() end,
				get = function() return db.texture end,
				validate = surf:List(),
				order = 2,
			},
			scale = {
				type = "group",
				name = "Scaling",
				desc = "Scale certain types of frames",
				order = 3,
				args = {
					full = {
						type = "range",
						name = "Scale full frames.",
						desc = "Player and Target",
						set = function(v) db.scale.full = v UpScale("full") end,
						get = function() return db.scale.full end,
						min = 0.2, max = 3.0, step = 0.02,
						order = 1,
					},
					medium = {
						type = "range",
						name = "Scale medium frames.",
						desc = "Pet and Party",
						set = function(v) db.scale.medium = v UpScale("medium") end,
						get = function() return db.scale.medium end,
						min = 0.2, max = 3.0, step = 0.02,
						order = 2,
					},
					small = {
						type = "range",
						name = "Scale small frames.",
						desc = "TargetOfTarget and Focus",
						set = function(v) db.scale.small = v UpScale("small") end,
						get = function() return db.scale.small end,
						min = 0.2, max = 3.0, step = 0.02,
						order = 3,
					},
				},
			},
			model = {
				type = "toggle",
				name = "Toggle 3D Models",
				desc = "Toggle showing 3D/2D portraits",
				set = function() self:ToggleModels() end,
				get = function() return db.showmodels end,
				order = 4,
			},
			deficit = {
				type = "toggle",
				name = "Show health deficit",
				desc = "Show HP deficit instead of percent",
				set = function(v) db.deficit = v self:UpAll() end, 
				get = function() return db.deficit end, 
				order = 6,
			},
			lock = {
				type = "toggle",
				name = "Unlock",
				desc = "Unlock the frames so you can move them",
				set = function() self:ToggleDrag() end, 
				get = function() return self.unlocked end, 
				order = 7,
			},
		},
	}
	
  	SlashCmdList["STELLARUF"] = function() 
  		dew:Open( UIParent,
		        'children', function() dew:FeedAceOptionsTable(opts) end,
		        'cursorX', true, 'cursorY', true ) 
  	end
   	SLASH_STELLARUF1 = "/stellaruf"

end



-- hide and unregister blizz defaults
function StellarUF:HideBlizzStuff()
	PlayerFrame:UnregisterAllEvents()
	PlayerFrameHealthBar:UnregisterAllEvents()
	PlayerFrameManaBar:UnregisterAllEvents()
	PlayerFrame:Hide()
	TargetFrame:UnregisterAllEvents()
	TargetFrame:Hide()
	TargetofTargetFrame:UnregisterAllEvents()
	TargetofTargetFrame:Hide()
	ComboFrame:UnregisterAllEvents()
	for i=1,4 do
		_G["PartyMemberFrame"..i]:UnregisterAllEvents()
		_G["PartyMemberFrame"..i.."PetFrame"]:UnregisterAllEvents()
		_G["PartyMemberFrame"..i]:Hide()
	end
end

-- cheap way to get support distances
function StellarUF:InitSpellRange()
	if self.cls == "PALADIN" then
		s30 = "Purify"
		s40 = "Holy Light"
	elseif self.cls == "PRIEST" then
		s30 = "Cure Disease"
		s40 = "Lesser Heal"
	elseif self.cls == "DRUID" then
		s30 = "Cure Poison"
		s40 = "Healing Touch"
	elseif self.cls == "MAGE" then
		s30 = "Remove Lesser Curse"
	elseif self.cls == "SHAMAN" then
		s30 = "Cure Poison"
		s40 = "Healing Wave"
	end
end


-- Events and processing functions

function StellarUF:RegisterEvents()
	-- units
	self:RegisterEvent("PLAYER_TARGET_CHANGED", "UpTarget")
	self:RegisterEvent("PLAYER_FOCUS_CHANGED", "UpFocus")
	self:RegisterBucketEvent("UNIT_PET", 0.5, "UpPet")
	self:ScheduleRepeatingEvent("tt", self.UpMetroUnits, 0.2, self)

	-- statusbars
	self:RegisterBucketEvent({ "UNIT_HEALTH", "UNIT_MAXHEALTH", }, 0.12, "UpHealth")
	self:RegisterBucketEvent({ "UNIT_MANA", "UNIT_MAXMANA", "UNIT_RAGE", "UNIT_ENERGY", }, 0.12, "UpPower")
	self:RegisterBucketEvent("UNIT_FOCUS", 0.3, "UpPower")
	self:RegisterEvent("UNIT_DISPLAYPOWER", "UpPowerType")
	
	-- text
	self:RegisterBucketEvent("RAID_ROSTER_UPDATE", 2, "UpGroupText")
	self:RegisterEvent("UNIT_COMBAT", "UpCombatStrings")
	self:RegisterEvent("UNIT_SPELLMISS", "UpCombatStrings")
	self:ScheduleRepeatingEvent("combat", self.UpCombatText, 0.1, self) -- fader
	
	-- icons
	self:RegisterBucketEvent("UNIT_AURA", 0.5, "UpAura")
	self:RegisterBucketEvent("RAID_TARGET_UPDATE", 2, "UpRaidTargetIcon")
	self:ScheduleRepeatingEvent("status", self.UpStatusesSlow, 1.0, self) -- I'm lazy
	self:ScheduleRepeatingEvent("status2", self.UpStatusesFast, 0.5, self)

	-- portrait
	self:RegisterBucketEvent("UNIT_PORTRAIT_UPDATE", 0.3, "UpPortrait")
	
	self:RegisterEvent("AceEvent_FullyInitialized", function()
		self:UpAll()
		self:RegisterEvent("PLAYER_ENTERING_WORLD", "UpAll")
	end)
end

-- some utility functions
--
local function IsInParty()
	return nparty() > 0
end
local function IsInRaid()
	return nraid() > 0
end
local function IsInGroup()
	return IsInParty() or IsInRaid()
end
local function IsUpable(unit, frame)
	return frame and exist(unit) and frame:IsShown()
end
local function ClassColorThis(text, cls)
	local c = RAID_CLASS_COLORS[cls or "blah"]
	if not c then 
		return text 
	end
	return format("|cff%02x%02x%02x%s|r", c.r*255, c.g*255, c.b*255, text)
end

---------------------------------------------------
--- *** GENERAL UPDATES *** ---
--

-- create party baseframes
function StellarUF:CreateParty()
	if not IsInGroup() then return end
	if self:IsEventRegistered("PARTY_MEMBERS_CHANGED") then 
		self:UnregisterEvent("PARTY_MEMBERS_CHANGED") 
	end
	if incombat() then 
		self:ScheduleLeaveCombatAction("CreateParty") 
		ChatFrame1:AddMessage("StellarUF: Party frames will be created after combat is over.")
		return 
	end
	for i=1,4,1 do
		local p, pp = format("party%d", i), format("partypet%d", i)
		self:CreateUnitFrame(p, "medium")
		self:CreateUnitFrame(pp, "tiny")
		su[pp]:ClearAllPoints()
		su[pp]:SetParent(su[p])
		su[pp]:SetPoint("TOPLEFT", su[p], "BOTTOMLEFT", -6, -1)
	end

	self:RegisterBucketEvent("PARTY_MEMBERS_CHANGED", 1, "UpParty")
	self:UpParty()
	partyvisible = true
end

local function ShowParty()
	for i=1,4,1 do
		local p, pp = format("party%d", i), format("partypet%d", i)
		RegisterUnitWatch(su[p], false)
		RegisterUnitWatch(su[pp], false)
	end
	partyvisible = true
end
local function HideParty()
	for i=1,4,1 do
		local p, pp = format("party%d", i), format("partypet%d", i)
		UnregisterUnitWatch(su[p])
		UnregisterUnitWatch(su[pp])
		su[p]:Hide()
		su[pp]:Hide()
	end
	partyvisible = false
end

-- hide party interface in raid if default option is checked
function StellarUF:RaidPartyUpdate()
	if not su.party1 or incombat() then return end

	if HIDE_PARTY_INTERFACE == "1" and IsInRaid() then
		if partyvisible then 
			HideParty() 
		end
	elseif not partyvisible then
		ShowParty()
	end
end

-- update target
function StellarUF:UpTarget()
	PlaySound("igCreatureNeutralSelect") -- they all sound the same to me
	self:UpUnit("target", true)
	
	-- reset fade effects
	local p = tar.portrait
	p.combattext.alpha = 0
	p.combattext:SetAlpha(0)
	p.glow.alpha = 0
	p.glow.duration = 0
	p.glow:SetAlpha(0)
end

-- update focus unit
function StellarUF:UpFocus()
	self:UpUnit("focus")
end

-- update targettarget and focustarget
function StellarUF:UpMetroUnits()
	self:UpUnit("targettarget") 
	self:UpUnit("focustarget") 
end

-- update pet bucket
function StellarUF:UpPet(units)
	for u in pairs(units) do 
		self:UpUnit(u)
	end
end

-- update all party members
function StellarUF:UpParty()
	for u,f in pairs(su) do
		if strfind(u, "party") then
			self:UpUnit(u)	
		end
	end
end

-- update everything for a certain unit
function StellarUF:UpUnit(unit, resetFade)
	local f = su[unit]
	if not IsUpable(unit, f) then return end
	
	self:UpUnitHealth(unit, f, resetFade)
	self:UpPowerType(unit)
	self:UpNameText(unit, f)
	self:UpInfoText(unit, f)
	self:UpPVPIcon(unit, f)
	self:UpStatusIcon(unit, f)
	self:UpClassIcon(unit, f)
	self:UpPartyIcons(unit, f)
	self:UpUnitPortrait(unit, f)
	self:UpUnitAura(unit, f)
	self:UpStatusText(unit, f)
	if unit == "target" then 
		self:UpRaidTargetIcon() 
		self:UpGroupText()
	end
end

-- update everything for all units
function StellarUF:UpAll()
	for u,f in pairs(su) do
		self:UpUnit(u, true)
	end
end

-- update status related elements (dead, offline, etc)
function StellarUF:UpStatusesSlow()
	for u,f in pairs(su) do
		if IsUpable(u,f) then
			self:UpInfoText(u,f)
			self:UpPVPIcon(u,f)
			self:UpPartyIcons(u,f)
		end
	end
end

function StellarUF:UpStatusesFast()
	for u,f in pairs(su) do
		if IsUpable(u,f) then
			self:UpStatusText(u,f)
			self:UpNameText(u,f)
			self:UpStatusIcon(u,f)
		end
	end
end

---------------------------------------------------
--- *** STATUSBARS *** ---
--

-- update health with bucket support
function StellarUF:UpHealth(units)
	for u in pairs(units) do
		self:UpUnitHealth(u, su[u])
	end
end

-- use this for changing bar values (adopted from DUF)
--   this way allows for a reverted status bar
local function SetBarValue(parent, value, reverse, hp, resetFade)
	value = min(value, 1)
	local b, bb, full = parent.bar, parent.backbar, parent:GetWidth()
	local old = max(b:GetWidth(), 0.1)
	local new = max(full * value, 0.1)
	
	-- bar loss fading
	if bb and not resetFade and old - new > 0.5 then
		bb:SetWidth(old)
		bb.alpha = 1.4
		bb:SetAlpha(1)
		if not reverse then
			bb:SetTexCoord(0, old / full, 0, 1)
		else
			bb:SetTexCoord(old / full, 0, 0, 1)
		end
	end
	b:SetWidth(new)
	
	-- set value of the main bar
	if not reverse then 
		b:SetTexCoord(0, max(value,0.001), 0, 1)
	else 
		b:SetTexCoord(max(value,0.001), 0, 0, 1)
	end
	
	-- hp green to red coloring
	if hp then
		local c = self.hpcolor
		local r,g
		if value > 0.5 then
			r = c.r + (((1 - value) * 2) * (c.g - c.r))
			g = c.g
		else
			r = c.g
			g = c.g - ((0.5 - value) * 2 * c.g)
		end
		b:SetVertexColor(r, g, c.b)
	end
end

local hpformat = "%d%%\n%s/%s"
local mpformat = "%s/%s"
local percformat = "%d%%"
local deficitformat = "-%s\n%s/%s"
local function BigNum(n)
	return (n > 100000 and format("%dk", n/1000)) or n
end
-- health text handler
local function SetHealthText(unit, perc, current, total)
	local f = su[unit]
	if f.frametype == "small" or f.frametype == "tiny" then
		f.hp.hptext:SetText(format(percformat, perc))
	else
	 	if MobHealth3 and unit == "target" then
	 		current, total = MobHealth3:GetUnitHealth("target", current, total, UnitName("target"), UnitLevel("target"))
	 	end
		if db.deficit then
			f.hp.hptext:SetText(format(deficitformat, BigNum(total - current), BigNum(current), BigNum(total)))
		else
			f.hp.hptext:SetText(format(hpformat, perc, BigNum(current), BigNum(total)))
		end
	end
end

-- update health
local uh, uhm = UnitHealth, UnitHealthMax
function StellarUF:UpUnitHealth(unit, f, resetFade)
	if not IsUpable(unit, f) then return end
	
	local current, total = uh(unit), uhm(unit)
	local perc = min(current/total, 1)
		
	if dead(unit) or not online(unit) then
		SetBarValue(f.hp, 0, f.reversed, true, resetFade)
		SetHealthText(unit, perc * 100, current, total)
	else			
		SetBarValue(f.hp, perc, f.reversed, true, resetFade)
		SetHealthText(unit, max(perc * 100, 1), current, total)
	end
end


-- update power with bucket support
function StellarUF:UpPower(units)
	for u in pairs(units) do
		self:UpUnitPower(u, su[u])
	end
end

-- update power value
local um, umm = UnitMana, UnitManaMax
function StellarUF:UpUnitPower(unit, f, resetFade)
	if not IsUpable(unit, f) then return end
	
	local current, total = um(unit), umm(unit)
	local perc = min(current/total, 1)
	local t = f.mp.mptext
	
	if dead(unit) or not online(unit) or total < 0.5 then
		SetBarValue(f.mp, 0, f.reversed, false, resetFade)
		if t then
			t:SetText(format(mpformat, 0, BigNum(total)))
		end
	else
		SetBarValue(f.mp, perc, f.reversed, false, resetFade)
		if t then
			t:SetText(format(mpformat, BigNum(current), BigNum(total)))
		end
	end
end

local powercolor = {
	[0] = { r=0.1, g=0.1, b=0.7, }, 		-- mana
	[1] = { r=0.5, g=0.1, b=0.1, }, 		-- rage
	[2] = { r=0.8, g=0.4, b=0, }, 		-- focus
	[3] = { r=0.6, g=0.6, b=0, }, 		-- energy
	[4] = { r=0.5, g=0.5, b=0, }, 		-- happiness
}
function StellarUF:UpPowerType(unit)
	local f = su[unit]
	if not IsUpable(unit, f) then return end
	
	local pc = powercolor[UnitPowerType(unit) or 0]
	local r,g,b = pc.r, pc.g, pc.b
	f.mp.bar:SetVertexColor(r, g, b)
	f.mp:SetBackdropColor(r/3, g/3, b/3, 0.6)
	self:UpUnitPower(unit, f, true)
end

---------------------------------------------------
--- *** PORTRAIT *** ---
--

function StellarUF:UpPortrait(units)
	for u in pairs(units) do
		if IsUpable(u, su[u]) then
			self:UpUnitPortrait(u, su[u])
		end
	end
end

-- update portrait
function StellarUF:UpUnitPortrait(unit, f)
	self:SetPortraitUnit(unit, f)
end

function StellarUF:SetPortraitUnit(unit, f)
	local p = f.portrait
	if visible(unit) and db.showmodels and p.d3 then
		if unit == "target" then
			p.d2:Hide()
		end
		p.d3:SetUnit(unit)
		p.d3:Show()
		p.d3:RefreshUnit()
		p.d3:SetCamera(0)
		p:SetBackdropColor(0.1, 0.1, 0.1, 0.8)
		if strfind(unit, "party") then
			SetPortraitTexture(p.d2, unit)
			p.d2:SetAlpha(0.3)
		end
	else
		if unit == "target" then 
			p.d2:Show() 
		end
		if p.d3 then 
			p.d3:Hide() 
		end
		SetPortraitTexture(p.d2, unit)
		p:SetBackdropColor(0.1, 0.1, 0.1, 0)
		p.d2:SetAlpha(1)
	end
end

function StellarUF:ToggleModels()
	if not db.showmodels then
		db.showmodels = true
		for u,f in pairs(su) do
			local p = f.portrait
			if p.d3 then
				p.d3:Show()
				p:SetBackdropColor(0.1, 0.1, 0.1, 0.8)
			end
			p.d2:SetAlpha(0.4)
			self:SetPortraitUnit(u, f)
		end
		su.player.portrait.d2:Hide()
		if exist("target") and visible("target") then 
			su.target.portrait.d2:Hide() 
		end
	else
		db.showmodels = false
		for u,f in pairs(su) do
			local p = f.portrait
			if p.d3 then
				p.d3:Hide()
				p:SetBackdropColor(0.1, 0.1, 0.1, 0)
			end
			p.d2:Show()
			self:SetPortraitUnit(u, f)
			p.d2:SetAlpha(0.9)
		end
	end
end

---------------------------------------------------
--- *** TEXT *** ---
--

-- update unit_combat events, based on DUF's code
function StellarUF:UpCombatStrings()
	local f = su[arg1]
	if not IsUpable(arg1, f) or not f.portrait.combattext then return end
	
	local color, value, symbol = "", "", ""
	if arg2 == "IMMUNE" then
		color = "|cFFFFFF00"
		value = CombatFeedbackText["IMMUNE"]
	elseif arg2 == "WOUND" then
		if arg3 == "ABSORB" then 
			color = "|cFFFFFFFF"
			value = TEXT(ABSORB)
		elseif arg4 ~= 0 then
			color = "|cFFFF0000-"
			value = arg4
			if arg3 == "CRITICAL" then 
				symbol = "!"
			elseif arg3 == "CRUSHING" then 
				symbol = "(" 
			end
			if arg4 > UnitLevel("player")*5 and f.portrait.glow then
				f.portrait.glow.duration = 2
				f.portrait.glow:SetVertexColor(1.0, 0.3, 0.3)
			end
		else
			color = "|cFFFFFF00"
			value = CombatFeedbackText["MISS"]
		end
	elseif arg2 == "BLOCK" then
		color = "|cFFFFFF00"
		value = TEXT(BLOCK)
	elseif arg2 == "HEAL" then
		color = "|cFF00FF00+"
		value = arg4
		if arg3 == "CRITICAL" then
			symbol = "!"
		end
		if arg4 > UnitLevel("player")*5 and f.glow then
			f.portrait.glow.duration = 2
			f.portrait.glow:SetVertexColor(0.3, 1.0, 0.3)
		end
	else
		value = CombatFeedbackText[arg2]
	end
	f.portrait.combattext:SetText(format("%s%s%s", color, value, symbol))
	f.portrait.combattext.alpha = 2.5
end

-- update fading of combat text
function StellarUF:UpCombatText()
	for u,f in pairs(su) do
		local ct = f.portrait.combattext
		if IsUpable(u, f) and ct then
			ct.alpha = max(ct.alpha - 0.1, 0)
			ct:SetAlpha(min(ct.alpha, 1))
		end
	end
end

-- Name formatting, based on blizz's code
local reactclr = {
	{ 0.85, 0, 0 },		-- 1 Hostile
	{ 0.85, 0, 0 },		-- 2 Hostile
	{ 0.85, 0.5, 0 },		-- 3 Unfriendly
	{ 0.85, 0.85, 0 },		-- 4 Neutral
	{ 0, 0.95, 0 },		-- 5 Friendly
	{ 0, 0.95, 0 },		-- 6 Friendly
	{ 0, 0.95, 0 },		-- 7 Friendly
	{ 0.97, 0.97, 0.97 },	-- 8 Something white
	{ 0.4, 0.4, 0.4 },		-- 9 Gray
}
white = reactclr[8]
function StellarUF:UpNameText(unit, f)
	if not f.hp.nametext then return end
	-- makes the 3d portrait slightly transparent if dead
	if f.portrait.d3 then 
		if dead(unit) then 
			f.portrait.d3:SetAlpha(0.6) 
		else 
			f.portrait.d3:SetAlpha(1) 
		end
	end
	
	local c
	local dark = 1
	f.hp.nametext:SetText(GetUnitName(unit))
	
	-- darken name if unit is out of range
	if friend(unit, "player") then
		if unit == "player" or (s30 and IsSpellInRange(s30, unit) == 1) then
			dark = 1
		elseif s40 and IsSpellInRange(s40, unit) == 1 then
			dark = 0.85
		elseif not s30 and not s40 and visible(unit) then
			dark = 1
		else
			dark = 0.7
		end
	end
	if unit == "player" or strfind(unit, "party") or unit == "pet" then
		c = white
	elseif UnitPlayerControlled(unit) then
		if UnitCanAttack(unit, "player") then -- Hostile players are red
			if not UnitCanAttack("player", unit) then
				c = white
			else
				c = reactclr[2]
			end
		elseif UnitCanAttack("player", unit) then -- Players we can attack but which are not hostile are yellow
			c = reactclr[4]
		elseif UnitIsPVP(unit) and not UnitIsPVPSanctuary(unit) and not UnitIsPVPSanctuary("player") then -- Players we can assist but are PvP flagged are green
			c = reactclr[6]
		else
			c = white
		end
	elseif UnitIsTapped(unit) and not UnitIsTappedByPlayer(unit) then
		c = reactclr[9]
	else
		c = reactclr[UnitReaction(unit, "player") or 8]
	end
	
	f.hp.nametext:SetTextColor(c[1]*dark, c[2]*dark, c[3]*dark)
end

-- level, creature type text
	-- elite/rare text
local classifytext = {
	worldboss = " Boss",
	rareelite = "+ |cffff8888Rare|r",
	elite = "+",
	rare = " |cffff8888Rare|r",
	normal = "",
}
function StellarUF:UpInfoText(unit, f)
	if not f.hp.infotext then return end
	
	local level = UnitLevel(unit)
	local lcls, cls = UnitClass(unit)
	-- color the target's level based on difficulty
	if unit == "target" then
		local hexcolor
		if level == -1 then 
			level = 999 
		end
		if friend(unit, "player") then
			hexcolor = "|cffeeeeee"
		else
			local lc = GetDifficultyColor(level)
			hexcolor = format("|cFF%02x%02x%02x", lc.r * 255, lc.g * 255, lc.b * 255)
		end
		if level == 999 then 
			level = "??" 
		end
		level = format("%s%s%s|r", hexcolor, level, classifytext[UnitClassification(unit) or "normal"])
			
		-- add creature type and color it by class if unit is player
		local creaturetype = UnitCreatureType(unit) or (UnitIsPlayer(unit) and "Humanoid") or "Unknown"
		if creaturetype == "Not specified" then 
			creaturetype = "Unknown" 
		end
		if UnitIsPlayer(unit) then 
			creaturetype = ClassColorThis(creaturetype, cls) 
		end
		level = format("%s %s", level, creaturetype)
	elseif lcls then
		level = format("%s %s", level, ClassColorThis(lcls, cls))
	end
	
	f.hp.infotext:SetText(level)
end

-- raid group numbers
local foundplayer, foundtarget
function StellarUF:UpGroupText()
	if not IsInRaid() then
		pla.group:Hide()
		tar.group:Hide()
		return
	end
	
	foundplayer, foundtarget = nil, nil
	if not UnitInRaid("target") then 
		foundtarget = true
		tar.group:Hide()
	end
	for i=1,MAX_RAID_MEMBERS,1 do
		local name,_,subgroup,_,_,_,_,_,_ = GetRaidRosterInfo(i)
		if not name then break end
		if name == UnitName("player") then
			pla.group.text:SetText(subgroup)
			pla.group:Show()
			foundplayer = true
		end
		if name == UnitName("target") then
			tar.group.text:SetText(subgroup)
			tar.group:Show()
			foundtarget = true
		end
		if foundplayer and foundtarget then break end
	end
end

-- Dead, Offline, Tapped, Civilian text
function StellarUF:UpStatusText(unit, f)
	if not f.hp.statustext then return end
	local deadtext, offline, tapped
	if dead(unit) then 
		deadtext = "DEAD" 
		if f.portrait.glow then
			f.portrait.glow.duration = 1
			f.portrait.glow:SetVertexColor(0.6, 0.6, 0.6)
		end
	end
	if UnitIsPlayer(unit) and not online(unit) then
		if deadtext then 
			offline = "/OFFLINE" 
		else 
			offline = "OFFLINE" 
		end
	end
	if unit == "target" and UnitIsTapped(unit) and not UnitIsTappedByPlayer(unit) then
		tapped = "TAPPED"
		self:UpNameText("target", f)
	end
	f.hp.statustext:SetText(format("%s%s", deadtext or tapped or "", offline or ""))
end

---------------------------------------------------
--- *** ICONS *** ---
--

-- update aura with bucket support
function StellarUF:UpAura(units)
	for unit in pairs(units) do
		self:UpUnitAura(unit, su[unit])
	end
end

-- buff/debuff icons
local ub, ud = UnitBuff, UnitDebuff
function StellarUF:UpUnitAura(unit, f)
	if not IsUpable(unit, f) then return end
	-- look for a dispellable debuff
	if f.debuffalert then
		local fd = f.debuffalert 
		fd:Hide()
		if friend(unit, "player") then
			for i=1,40,1 do
				local debuff,_,icon,count,type = ud(unit, i, true)
				if not debuff then break end
				local dc = type ~= nil and DebuffTypeColor[type]
				if dc then
					fd.icon:SetTexture(icon)
					fd:SetBackdropColor(dc.r, dc.g, dc.b)
					fd:Show()
					if count > 1 then
						fd.count:SetText(count)
						fd.count:Show()
					else
						fd.count:Hide()
					end
					break
				end	
			end	
		end
	end
	
	-- don't handle buffs/debuffs for player
	if unit == "player" then return end
	-- update buffs
	local hasBuffs, hasMoreBuffs
	for i=1,32,1 do
		local button = f[format("buff%d", i)]
		if not button then break end
		local buff,_,icon,count = ub(unit, i)
		if buff then
			button.icon:SetTexture(icon)
			if count > 1 then
				button.count:SetText(count)
				button.count:Show()
			else
				button.count:Hide()
			end
			button:SetBackdropColor(0.1,0.1,0.1)
			button:Show()
			hasBuffs = true
			if i > 16 then hasMoreBuffs = true end
		else
			button:Hide()
		end
	end
	
	-- update debuffs
	for i=1,40,1 do
		local button = f[format("debuff%d", i)]
		if not button then break end
		local debuff,_,icon,count,type = ud(unit, i)
		if debuff then
			local c = DebuffTypeColor[type or "none"]
			button.icon:SetTexture(icon)
			button:SetBackdropColor(c.r, c.g, c.b)
			if count > 1 then
				button.count:SetText(count)
				button.count:Show()
			else
				button.count:Hide()
			end
			button:Show()
		else
			button:Hide()
		end
	end
		
	-- switch locations of debuffs if no buffs are shown
	if unit == "target" or unit == "targettarget" or unit == "focus" or unit == "focustarget" then
		if hasMoreBuffs and f.buff17 then -- extreme case
			f.debuff1:SetPoint("TOPLEFT", f.buff17, "BOTTOMLEFT", 0, -2)
		elseif hasBuffs then
			f.debuff1:SetPoint("TOPLEFT", f.buff1, "BOTTOMLEFT", 0, -2)
		else
			f.debuff1:SetPoint("TOPLEFT", f.buff1, "TOPLEFT", 0, 0)
		end
	end
end

-- status icon to show when an unit is offline, dead, incombat, or resting 
function StellarUF:UpStatusIcon(unit, f)
	if not f.statusicon then return end
	local icon = f.statusicon.texture
	if not online(unit) then
		icon:SetTexture("Interface\\CharacterFrame\\Disconnect-Icon")
		icon:SetTexCoord(0, 1, 0, 1)
		icon:Show()
	elseif dead(unit) then
		icon:SetTexture("Interface\\TargetingFrame\\UI-TargetingFrame-Skull")
		icon:SetTexCoord(0, 1, 0, 1)
		icon:Show()
	elseif UnitAffectingCombat(unit) then
		icon:SetTexture("Interface\\CharacterFrame\\UI-StateIcon")
		icon:SetTexCoord(.5625, .9, .08, .4375)
		icon:Show()
	elseif unit == "player" and IsResting() then
		icon:SetTexture("Interface\\CharacterFrame\\UI-StateIcon")
		icon:SetTexCoord(.0625, .4375, .0625, .4375)
		icon:Show()
		if f.portrait.glow then
			f.portrait.glow.duration = 2
			f.portrait.glow:SetVertexColor(0.4, 0.4, 1.0)
		end
	else
		icon:Hide()
	end
end

-- class icon
local clsicon = {
	WARRIOR = { .25, 0, 0, .25, },
	MAGE = { .5, .25, 0, .25, },
	ROGUE = { .75, .5, 0, .25, },
	DRUID = { 1, .75, 0, .25, },
	HUNTER = { .25, 0, .25, .5, },
	SHAMAN = { .5, .25, .25, .5, },
	PRIEST = { .75, .5, .25, .5, },
	WARLOCK = { 1, .75, .25, .5, },
	PALADIN = { .25, 0, .5, .75, },
}
function StellarUF:UpClassIcon(unit, f)
	if not f.classicon then return end
	local icon = f.classicon.texture
	local _,class = UnitClass(unit)
	if not class then
		icon:SetTexCoord(0, 1, 0, 1)
		icon:SetTexture("Interface\\Icons\\INV_Misc_QuestionMark")
		return
	end
	local c = clsicon[class] or clsicon["WARRIOR"]
	icon:SetTexture("Interface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes")
	icon:SetTexCoord(c[2] + 0.0175, c[1] - 0.0175, c[3] + 0.018, c[4] - 0.01)	
end

-- pvp flag icon
function StellarUF:UpPVPIcon(unit, f)
	if not f.pvpicon then return end
	local icon = f.pvpicon.texture
	if UnitIsPVPFreeForAll(unit) then 
		icon:SetTexture("Interface\\TargetingFrame\\UI-PVP-FFA")
		icon:Show()
	elseif UnitFactionGroup(unit) and UnitIsPVP(unit) then 
		icon:SetTexture(format("Interface\\TargetingFrame\\UI-PVP-%s", UnitFactionGroup(unit)))
		icon:Show()
	else 
		icon:Hide()
	end
end

-- leader/loot icon
function StellarUF:UpPartyIcons(unit, f)
	-- leader icon
	if not f.leadericon then return end
	if UnitIsPartyLeader(unit) then
		f.leadericon.texture:Show()
	else
		f.leadericon.texture:Hide()
	end

	-- loot icon
	local _, lootMaster = GetLootMethod()
	local looter
	if lootMaster == 0 then 
		looter = "player"
	elseif lootMaster then 
		looter = format("party%d", lootMaster) 
	end
	if looter and UnitIsUnit(looter, unit) then
		f.looticon.texture:Show()
	else
		f.looticon.texture:Hide()
	end
end

-- raid target symbols
function StellarUF:UpRaidTargetIcon()
	if not IsUpable("target", tar) then return end
	
	local iconID = GetRaidTargetIndex("target")
	if iconID then
		SetRaidTargetIconTexture(tar.raidtarget.texture, iconID)
		tar.raidtarget:Show()
	else
		tar.raidtarget:Hide()
	end 
end



